
;place this proto near the beginning of the source file

nArb2StrL PROTO   :DWORD,:DWORD,:DWORD,:DWORD,:DWORD

;-------------------------------------------------------------------------------------------

        OPTION  EPILOGUE:NONE
        OPTION  PROLOGUE:NONE

        ALIGN   16

nArb2StrL PROC lpBinary:DWORD,dwNumberOfBytes:DWORD,lpBuffer:DWORD,dwSize:DWORD,dwMode:DWORD
;
; CONVERT SIGNED OR UNSIGNED ARBITRARY INTEGER TO ASCII DECIMAL STRING
; LING LONG KAI FANG METHOD (HORNERS RULE) VERSION 1a
;
;Written by DednDave, September 2009
;Specialized code assistance by Drizz
;
;  When developing this routine, we started out by modifying some old 16-bit code
;written by Richard Pavlicek in 1998. That code is gone now and many improvements
;have been made. The old code used the "divide-and-conquer" method, but it did
;create 8 output bytes per pass. This version employs the "Ling Long Kai Fang"
;polynomial evaluation technique (modernly called "Horners Rule"). In some ways, it
;is like working the divide-and-conquer method in reverse. This method has a number
;of advantages; First, the input value need not be copied in order to be retained, as
;ling long kai fang does not modify the original. This also means better use of
;memory, allowing larger input values than an equivalent divide-and-conquer routine
;(this is only evident for extremely large values). Second, it should be a little bit
;faster, as we are trading DIV operations for MUL's. Some comparison testing is needed.
;A third advantage is that ling long kai fang lends itself very nicely to evaluating
;signed or unsigned integers. Divide-and-conquer has a penalty for negative values.
;
;  Error checking has been added. Several speed improvements have been made, although
;the routine is probably still not optimal. It is accurate and reliable and it can
;handle very large input values. This routine does not use any FPU or MMX/SSE
;instructions, leaving those registers free for other use.
;
;  Thanks go to Drizz for helping me speed up critical sections of code. These are the
;snippets of "multiplication by inverse to divide by constant" code. Kudos, Drizz.
;
;-----------------------------------------
;Usage:
;
;       INVOKE  nArb2StrL,
;               lpBinary,
;               dwNumberOfBytes,
;               lpBuffer,
;               dwSize,
;               dwMode
;
;        lpBinary = Offset of binary input value
; dwNumberOfBytes = Number of bytes in input value
;        lpBuffer = Offset of output buffer
;          dwSize = Size of the output buffer in bytes
;          dwMode = Signed/Unsigned (any negative value selects mode = signed)
;
;-----------------------------------------
;Returns:
;
; edx = Pointer to first byte of output string
; ecx = Length of output string, in bytes (may include "-" sign)
; eax = Return status:
;   eax = 0   ERROR_SUCCESS
;             "The operation completed successfully."
;          a) no errors:
;               the ASCII decimal string pointer and length are returned
;   eax = 87  ERROR_INVALID_PARAMETER
;             "The parameter is incorrect."
;          a) dwNumberOfBytes is 0, negative, or too large:
;               the specified output buffer pointer with length = 0 is returned
;          b) dwSize is 0, negative, or too large:
;               the specified output buffer pointer with length = 0 is returned
;   eax = 122 ERROR_INSUFFICIENT_BUFFER
;             "The data area passed to a system call is too small."
;             (windows list of error descriptions is poorly devised)
;          a) output buffer overflow:
;               the specified output buffer pointer with length = 0 is returned
;
;-----------------------------------------
;Notes:
;
; 1) The minimum output buffer size is 4 bytes.
;
; 2) To achieve optimal performance, alignment of the input value and output buffer
;    is critical. The input value should be 4-aligned. The byte immediately following
;    the last byte of the output buffer should also be 4-aligned.
;
; 3) The output string is right-justified in the output buffer field.
;
; 4) The input value is not altered during operation.
;
; 5) No verification is made to insure the address parameters are valid (i.e. access
;    privileges). However, if adding the output buffer address to its' size causes
;    overflow, the size is flagged as invalid. Similarly, if adding the input address
;    to the input length causes overflow, the length is flagged as invalid. The later
;    may cause confusion, as one might expect "ERROR_NOT_ENOUGH_MEMORY", instead.
;
;-----------------------------------------

;set up the stack frame

        push    ebx                  ;save the registers
        push    esi
        push    edi
        push    ebp
        mov     ebp,esp              ;ebp = stack frame pointer

;verify valid input length parameter

        mov     ecx,[ebp+24]         ;parm2 = input length in bytes
        xor     ebx,ebx              ;zero ebx
        cmp     ebx,ecx              ;error if length < +1
        jge     A2Str0               ;error - input length invalid

        mov     edi,ecx              ;input bytes in edi
        dec     edi                  ;count to last byte
        add     edi,[ebp+20]         ;parm1 = input value address
        jc      A2Str0               ;error - input length invalid

;verify valid output buffer size parameter

        mov     esi,[ebp+32]         ;parm4 = size of output buffer in bytes
        mov     bl,3                 ;buffer length test/adjust value
        cmp     ebx,esi              ;error if buffer size < +4
        jge     A2Str0               ;error - invalid output buffer size

        dec     esi                  ;count to last byte
        add     esi,[ebp+28]         ;parm3 = address of output buffer
        jc      A2Str0               ;error - invalid output buffer size

        sub     esi,ebx              ;point to last buffer dword
        push    esi                  ;[ebp-4] = offset of last output dword
        push    esi                  ;[ebp-8] = offset of first active dword

;create the signed/unsigned mode control mask

        mov     eax,[ebp+36]         ;mode control in eax
        cdq                          ;mode control mask in edx

;create the sign XOR mask

        mov     al,[edi]             ;high-order byte in al
        cbw                          ;sign extended into ax
        cwde                         ;sign extended into eax
        mov     al,ah                ;fill al with sign bits
        and     eax,edx              ;signed/unsigned mode mask
        mov     ebx,eax              ;and in ebx
        push    eax                  ;[ebp-12] = sign XOR mask

;reduce input length to ignore unused high-order bytes

        std                          ;scan downward
        repz    scasb                ;skip over unused bytes
        cld                          ;direction = up

;load and align the first (high-order) input dword

        and     cl,3                 ;0-3 in cl
        jz      A2Str1               ;load 1 byte

        cmp     cl,2                 ;test alignment
        jz      A2Str2               ;load 3 bytes

        jb      A2Str3               ;load 2 bytes

        dec     edi                  ;adjust edi
        dec     edi                  ;adjust edi
        mov     eax,[edi]            ;load 4 bytes
        jmp short A2Str4             ;correct for sign

;invalid input length parameter or invalid output size parameter

A2Str0: cld                          ;always return "up"
        mov     eax,87               ;return code in eax
        jmp     A2StrG               ;error exit

;load and align the first (high-order) input dword (cont.)

A2Str1: inc     edi                  ;adjust edi
        mov     al,[edi]             ;load 1 byte
        jmp short A2Str4             ;correct for sign

A2Str2: mov     ah,[edi+1]           ;load 3rd byte
        dec     edi                  ;adjust edi
        bswap   eax                  ;3rd byte in place

A2Str3: mov     ax,[edi]             ;load 2 bytes

;correct for sign and initialize stack frame constants

A2Str4: push    8461CEFDh            ;[ebp-16] = 8461CEFDh
        xor     eax,ebx              ;xor with sign mask
        push    0ABCC7711h           ;[ebp-20] = 0ABCC7711h
        push    100000000            ;[ebp-24] = 100,000,000
        xor     ecx,ecx              ;zero high dword
        push    55E63B89h            ;[ebp-28] = 55E63B89h

;test to see if the first dword is the only dword

        cmp     edi,[ebp+20]         ;edi = buffer base address ?
        jz      A2StrB               ;yes - do last dword

        xchg    eax,ecx              ;hi/lo swap for multiplication

;ling long kai fang outer loop

        ALIGN   16                   ;align the loop

A2Str5: xchg    al,cl                ;swap for multiplication
        push    edi                  ;save input pointer
        bswap   eax                  ;reorder for multiplication
        mov     edx,[ebp-16]         ;8461CEFDh partial multiplier
        mov     [esi],eax            ;store low dword
        ror     ecx,8                ;adjust high dword
        mul     edx                  ;partial multiply to divide
        mov     eax,[esi]            ;recall low dword
        mov     edi,edx              ;save partial result
        xor     ebx,ebx              ;zero for carry
        mov     edx,[ebp-20]         ;0ABCC7711h partial multiplier
        push    ebx                  ;save a zero for later
        mul     edx                  ;partial multiply to divide
        add     edi,eax              ;partial sum
        mov     eax,8461CEFDh        ;partial multiplier
        adc     ebx,edx              ;partial sum
        mul     ecx                  ;partial multiply to divide
        add     eax,edi              ;partial sum
        pop     edi                  ;recall zero in edi
        adc     ebx,edx              ;partial sum
        mov     eax,0ABCC7711h       ;partial multiplier
        adc     edi,edi              ;partial sum
        mul     ecx                  ;partial multiply to divide
        add     eax,ebx              ;complete the sum
        adc     edi,edx              ;complete the sum
        shrd    eax,edi,18           ;adjust the low quotient (carry)
        mov     edx,390625           ;remainder multiplier
        shr     edi,18               ;adjust the high quotient (carry)
        push    eax                  ;save the low quotient (carry)
        mul     edx                  ;extend the remainder
        neg     eax                  ;negate the difference
        add     eax,[esi]            ;subtract the difference
        shl     eax,8                ;pad the low byte
        mov     [esi],eax            ;store the remainder

;ling long kai fang inner loop

        ALIGN   16                   ;align the loop

A2Str6: sub     esi,8                ;next output dword
        pop     eax                  ;recall low carry dword
        cmp     esi,[ebp-8]          ;dword active yet ?
        jb      A2Str8               ;no - test carry

        mov     ecx,[esi]            ;yes - load output dword
        add     ecx,edi              ;add high carry dword

A2Str7: mov     edx,[ebp-16]         ;8461CEFDh partial multiplier
        mov     [esi],eax            ;store low dword
        mul     edx                  ;partial multiply to divide
        xor     ebx,ebx              ;zero for carry
        mov     edi,edx              ;save partial result
        push    [ebp-20]             ;0ABCC7711h partial multiplier
        mov     eax,[esi]            ;recall low dword
        pop     edx                  ;0ABCC7711h in edx
        push    ebx                  ;save a zero for later
        mul     edx                  ;partial multiply to divide
        push    [ebp-16]             ;8461CEFDh partial multiplier
        add     edi,eax              ;partial sum
        pop     eax                  ;8461CEFDh in eax
        adc     ebx,edx              ;partial sum
        mul     ecx                  ;partial multiply to divide
        add     eax,edi              ;partial sum
        pop     edi                  ;recall zero in edi
        adc     ebx,edx              ;partial sum
        mov     eax,0ABCC7711h       ;partial multiplier
        adc     edi,edi              ;partial sum
        mul     ecx                  ;partial multiply to divide
        add     eax,ebx              ;complete the sum
        adc     edi,edx              ;complete the sum
        shrd    eax,edi,26           ;adjust the low quotient (carry)
        mov     edx,[ebp-24]         ;100,000,000 remainder multiplier
        shr     edi,26               ;adjust the high quotient (carry)
        push    eax                  ;save the low quotient (carry)
        mul     edx                  ;extend the remainder
        sub     [esi],eax            ;subtract the difference
        jmp     A2Str6               ;next output dword

A2Str8: or      eax,eax              ;low carry dword ?
        jz      A2StrA               ;no - test high carry dword

A2Str9: cmp     esi,[ebp+28]         ;yes - test buffer base address
        jb      A2StrF               ;error - output buffer overflow

        mov     ecx,edi              ;init with high carry dword
        mov     [ebp-8],esi          ;update active dword pointer
        jmp     A2Str7               ;continue inner loop

A2StrA: or      edi,edi              ;high carry dword ?
        jnz     A2Str9               ;yes - activate output dword

;resume outer loop

        pop     edi                  ;recall input pointer
        sub     edi,4                ;next input dword
        mov     esi,[ebp-4]          ;last output dword
        mov     ecx,[edi]            ;load the input dword
        xor     ecx,[ebp-12]         ;xor with sign mask
        add     ecx,[esi]            ;add the output dword
        adc     eax,eax              ;extend the carry
        cmp     edi,[ebp+20]         ;input value base address
        ja      A2Str5               ;until the last input dword

        xchg    eax,ecx              ;hi/lo swap for multiplication

;last input dword

A2StrB: sub     eax,[ebp-12]         ;add one if negative
        sbb     ecx,[ebp-12]         ;allow it to carry
        shr     ecx,1                ;33 bit value divided as 32
        rcr     eax,1                ;into the low dword
        rcr     ecx,1                ;save the low bit in ecx
        mov     edx,[ebp-28]         ;55E63B89h = 2^56/50,000,000
        mov     edi,eax              ;save the dividend
        mul     edx                  ;multiply to divide
        shr     edx,24               ;adjust the quotient (carry)
        imul    eax,edx,50000000     ;extend the remainder
        sub     edi,eax              ;subtract the difference
        shl     ecx,1                ;recall the low bit
        rcl     edi,1                ;into the remainder

;last input dword loop

        ALIGN   16                   ;align the loop

A2StrC: mov     [esi],edi            ;store the output dword
        or      edx,edx              ;carry ?
        jz      A2StrH               ;no - convert to ASCII

        sub     esi,8                ;next output dword (pair)
        xchg    eax,edx              ;carry into dividend
        cmp     esi,[ebp-8]          ;dword active yet ?
        jb      A2StrE               ;no - activate it

        add     eax,[esi]            ;yes - add output dword

A2StrD: mov     edx,[ebp-28]         ;55E63B89h = 2^57/100,000,000
        mov     edi,eax              ;save the dividend
        mul     edx                  ;multiply to divide
        shr     edx,25               ;adjust the quotient (carry)
        imul    eax,edx,100000000    ;extend the remainder
        sub     edi,eax              ;subtract the difference
        jmp     A2StrC               ;store it and do next dword

A2StrE: cmp     esi,[ebp+28]         ;test buffer base address
        jb      A2StrF               ;error - output buffer overflow

        mov     [ebp-8],esi          ;update active dword pointer
        jmp     A2StrD               ;resume last input loop

;output buffer overflow

A2StrF: mov     eax,122              ;return code in eax

;error exit

A2StrG: xor     ecx,ecx              ;length = 0
        mov     edx,[ebp+28]         ;output buffer address
        jmp     A2StrN               ;exit the routine

;ASCII conversion

A2StrH: mov     ebx,10               ;digit multiplier in ebx
        mov     edi,[ebp-4]          ;pointer to last dword
        mov     esi,4294968          ;2^32/1000 in esi
        jmp short A2StrK             ;loop entry

;ASCII conversion loop

        ALIGN   16                   ;align the loop

A2StrJ: sub     edi,8                ;poin to next dword pair
        mov     [edi+4],ecx          ;store second dword of last pair

A2StrK: mov     eax,[edi]            ;load the next dword
        mov     edx,0D1B71759h       ;2^45/10,000
        mov     ecx,eax              ;save the dividend
        mul     edx                  ;multiply to divide
        shr     edx,13               ;adjust the quotient
        push    edx                  ;save the quotient
        imul    eax,edx,10000        ;extend the remainder
        neg     eax                  ;negate the difference
        add     eax,ecx              ;subtract the difference
        mul     esi                  ;extend the 1st digit
        mov     ecx,edx              ;digit 1 in cl
        mul     ebx                  ;extend the 2nd digit
        mov     ch,dl                ;digit 2 in ch
        mul     ebx                  ;extend the 3rd digit
        bswap   ecx                  ;digits 1 & 2 up high
        mov     ch,dl                ;digit 3 in ch
        mul     ebx                  ;extend the 4th digit into dl
        lea     ecx,[edx+ecx+'0000'] ;combine and make ASCII
        pop     eax                  ;recall the quotient
        bswap   ecx                  ;re-order bytes
        mov     [edi],ecx            ;store low 4 digits
        mul     esi                  ;extend the 1st digit
        mov     ecx,edx              ;digit 1 in cl
        mul     ebx                  ;extend the 2nd digit
        mov     ch,dl                ;digit 2 in ch
        mul     ebx                  ;extend the 3rd digit
        bswap   ecx                  ;digits 1 & 2 up high
        mov     ch,dl                ;digit 3 in ch
        mul     ebx                  ;extend the 4th digit into dl
        lea     ecx,[edx+ecx+'0000'] ;combine and make ASCII
        cmp     edi,[ebp-8]          ;last dword ?
        bswap   ecx                  ;re-order bytes
        ja      A2StrJ               ;no - continue

;test buffer overflow before expanding the last dword

        mov     ebx,edi              ;pointer to first dword
        mov     esi,[ebp+28]         ;output buffer base address
        sub     ebx,4                ;adjust for dword pair
        mov     ax,2D30h             ;'0' in al, '-' in ah
        cmp     ebx,esi              ;room to expand it ?
        jb      A2StrP               ;no - special loop

        mov     edi,ebx              ;yes - point to it
        mov     [ebx],ecx            ;and store it

;suppress leading zeros

A2StrL: mov     ecx,8                ;only need to test 8
        xor     edx,edx              ;zero for later
        repz    scasb                ;skip over leading zeros
        dec     edi                  ;point to first non-zero

;set the minus sign byte

        cmp     edx,[ebp-12]         ;negative value ?
        jz      A2StrM               ;no - skip the minus sign

        dec     edi                  ;yes - point to it
        cmp     edi,esi              ;buffer overflow ?
        jb      A2StrF               ;error - output buffer overflow

        mov     [edi],ah             ;store the minus sign

;set return values

A2StrM: mov     ecx,[ebp-4]          ;offset of last dword
        xchg    eax,edi              ;edx = offset of ascii string
        add     ecx,4                ;end of string
        xchg    eax,edx              ;eax = 0 if no error
        sub     ecx,edx              ;ecx = ascii string length

;restore registers and exit

A2StrN: leave                        ;mov esp,ebp - pop ebp
        pop     edi                  ;restore the registers
        pop     esi
        pop     ebx
        ret     20                   ;return and discard 5 dwords

;allow single byte placement of last dword if required

A2StrP: bswap   ecx                  ;switch order back

A2StrQ: cmp     ecx,'0000'           ;all done ?
        jz      A2StrL               ;yes - suppress leading zeros

        dec     edi                  ;no - point to next byte
        cmp     edi,esi              ;buffer overflow ?
        jb      A2StrF               ;error - output buffer overflow

        mov     [edi],cl             ;store the byte
        shrd    ecx,eax,8            ;get the next one in cl
        jmp     A2StrQ               ;loop until done

nArb2StrL ENDP

        OPTION  EPILOGUE:EpilogueDef
        OPTION  PROLOGUE:PrologueDef

;-------------------------------------------------------------------------------------------
